"""
Author: Peter Mawhorter
Date: 2023-8-9
Purpose: Script for auto-downloading and organizing CS 111 materials:
    project starter code, exercise notebooks, lecture notebooks, lecture
    slides, and sometimes lab starter code. Looks at the calendar page to
    find links, and then follows those links to look for download links,
    which it compares against what's already been downloaded. It doesn't
    check things in detail: if a download folder exists, it assumes
    the download is up-to-date and/or has already been worked on, and so
    it only downloads stuff that's totally new.
"""

# Version number for this script
__version__ = "0.1"

import os
import zipfile
import html.parser
import urllib.request
import urllib.parse

# Whether to re-scan all links or to pay attention to the skip list to
# speed things up:
RESCAN_ALL = False

# Current semester (affects calendar URL and project links)
SEMESTER = "spring24"

# URL for the calendar page to scrape
CALENDAR_URL = (
    f"https://cs111.wellesley.edu/archive/cs111_{SEMESTER}"
    f"/public_html/schedule"
)

SITE_LOC = urllib.parse.urlparse(CALENDAR_URL).netloc

# Format for project starter file links
PROJECT_STARTER_FMT = (
    "https://cs.wellesley.edu/potluck/cs111/{semester}/starter/{task}.zip"
)


class ScrapingParser(html.parser.HTMLParser):
    """
    As it parses, creates a 'result' value which is a list of triples
    containing a section hierarchy level number (e.g., 1 for 'h1' and 2
    for 'h2'), followed by a title for that section (accumulated without
    HTML from within the heading tag), followed by a list of link URLs
    that occurred within that section (i.e., before the next
    equivalent-or-higher-level heading) Includes multiple copies of
    links when they appear under multiple levels of heading hierarchy.

    In the resulting dictionary, links outside of any headings appear in
    a tuple with level 0 and title ''. This entry will always be present
    and will contain all links in the document.

    Links that appear within a header tag will NOT count as being inside
    that tag.

    Sections start when their header tag is closed, so nested header
    tags won't apply beyond the close of the outer section, given that
    starting a higher-level section ends all lower-level sections. Who
    would use nested header tags anyways? T_T

    An example result might look like:

    ```py
    [
        (
            0,
            '',
            [
                'www.example.com/outside',
                'www.example.com',
                'www.example.com/another',
            ]
        ),
        (1, 'Page Title', ['www.example.com', 'www.example.com/another']),
        (2, 'Section 1': ['www.example.com']),
        (2, 'Section 2': ['www.example.com/another']),
        (1, 'Another': []),
        (1, 'Another': [])
    ]
    ```
    """
    def __init__(self):
        "Initializes variables then calls superclass init."
        self.result = [(0, '', [])]
        self.expecting = []  # which tag end(s) are we expecting
        self.titles = []  # accumulators for tag titles
        super().__init__()

    def catalogueLink(self, url):
        """
        Adds the given URL to the appropriate heading lists, based on
        the current result value.
        """
        nextLevel = self.result[-1][0]
        index = len(self.result) - 1
        while nextLevel >= 0 and index > 0:
            levelHere, titleHere, urlsHere = self.result[index]
            if levelHere <= nextLevel:
                urlsHere.append(url)
                nextLevel = levelHere - 1
            index -= 1

    def handle_starttag(self, tag, attrs):
        """
        Sets up processing for heading tags; grabs URLs from links.
        """
        if tag in ('h1', 'h2', 'h3', 'h4', 'h5', 'h6'):
            self.expecting.append(tag)
            self.titles.append('')
        elif tag == 'a':
            href = None
            for attr, val in attrs:
                if attr == 'href':
                    href = val
                    break
            if href is not None:
                self.catalogueLink(href)

    def handle_endtag(self, tag):
        """
        Finalizes heading tags, now that their titles are known.
        """
        if tag in ('h1', 'h2', 'h3', 'h4', 'h5', 'h6'):
            level = int(tag[1:])
            exp = self.expecting.pop()
            if tag != exp:
                raise ValueError(
                    f"Malformed HTML: got {tag} end while expecting an"
                    f" end for {exp}."
                )

            title = self.titles.pop()  # grab accumulated title
            self.result.append((level, title, []))

    def handle_data(self, data):
        """
        Accumulates title values if we're inside of a heading tag.
        """
        if len(self.titles) > 0:
            for level in range(len(self.titles)):  # at each level
                self.titles[level] += data  # accumulate title


def scrapeLinksBySection(url):
    """
    Given a page URL, loads that page and returns a list of (level,
    title, links) tuples for each section in the page (see
    `ScrapingParser`).
    """
    parser = ScrapingParser()

    with urllib.request.urlopen(url) as htmlInput:
        parser.feed(htmlInput.read().decode('utf-8'))

    return parser.result


def folderNameForURL(url):
    """
    Returns the folder name that a download with a given URL should be
    stored under, after being unzipped. Returns None for URLs that don't
    match one of the recognized patterns.
    """
    if (
        url.startswith('/lectures/')
     or url.startswith('/labs/')
     or url.startswith('/exercises/')
     or url.startswith('/projects/')
    ):
        return url[1:].rstrip('/ ')
    else:
        return None


def linkDest(folder, link):
    """
    Returns a URL, destination pair for the given link and folder to put
    it in. For .zip URLs, uses just the folder, since that will be the
    extraction target.
    """
    linkInfo = urllib.parse.urlparse(link)
    path = linkInfo.path
    lastPart = path.split('/')[-1]
    unquoted = urllib.parse.unquote(lastPart)
    return (link, folder + '/' + unquoted)


def collectLink(addTo, folder, baseURL, link):
    """
    Given a result list, a folder destination, a base URL, and a link,
    adds a (URL, destination) pair to the result list for the given
    link, but skips fragment-only and external links.
    """
    if link.startswith('#'):
        return
    link = urllib.parse.urljoin(baseURL, link)
    info = urllib.parse.urlparse(link)
    if info.netloc != SITE_LOC:
        return
    addTo.append(linkDest(folder, link))


def findLectureURLs(folder, baseURL):
    """
    Given a lecture folder path string and a lecture link URL, loads
    that page and finds the download links for the slides and the
    lecture notebooks, returning a list of (URLs, download-as) pairs.
    """
    result = []
    sections = scrapeLinksBySection(baseURL)
    for level, title, links in sections:
        title = title.strip()
        if level == 2 and title in ('Slides', 'Code'):
            for link in links:
                # don't process any links after the solutions reveal link
                if link.strip() == '#solutions':
                    break
                collectLink(result, folder, baseURL, link)

    return result


def findLabURLs(folder, baseURL):
    """
    Given a lab link URL, loads that page and finds the download link
    for the starter files, if there is one. Returns a list of URL,
    destination pairs.
    """
    result = []
    sections = scrapeLinksBySection(baseURL)
    for level, title, links in sections:
        if title.strip() in (
            "Ready? Let's grab the starter code for today's lab",
            'Lab Setup'
        ):
            for link in links:
                collectLink(result, folder, baseURL, link)

    return result


def findExercisesURLs(folder, baseURL):
    """
    Given an exercises link URL, loads that page and finds the download
    links for the zip files for that week's exercises, returning a list
    of URL, destination pairs.
    """
    result = []
    sections = scrapeLinksBySection(baseURL)
    for level, title, links in sections:
        if title.strip() == 'Code':
            for link in links:
                collectLink(result, folder, baseURL, link)

    return result


def findProjectURLs(folder, baseURL):
    """
    Given a project link URL, loads that page and finds the task IDs, so
    that it can construct the download links for each task. Returns a
    list of URL, destination pairs.
    """
    result = []
    sections = scrapeLinksBySection(baseURL)
    for level, title, links in sections:
        if level == 2 and title.strip() == 'Tasks':
            for link in links:
                info = urllib.parse.urlparse(link)
                if (
                    info.path
                and 'projects/instructions' in info.path
                ):
                    taskID = info.path.rstrip('/').split('/')[-1]
                    result.append(
                        linkDest(
                            folder,
                            PROJECT_STARTER_FMT.format(
                                semester=SEMESTER,
                                task=taskID
                            )
                        )
                    )

    return result


def main():
    """
    Loads the calendar page (via the `CALENDAR_URL`) and looks for links
    to lectures, labs, exercises, and projects (using a very fragile
    and simple line-based method).

    For each link found in each category, uses typeAndFolderName to
    deduce
    """
    sections = scrapeLinksBySection(CALENDAR_URL)

    downloadMap = {}  # map from URLs to folders

    if os.path.exists('skip.txt'):
        with open('skip.txt', 'r') as inputFile:
            skip = [line.strip() for line in inputFile.readlines()]
    else:
        skip = []

    print("Finding calendar links...")
    for level, title, links in sections:
        if title not in ("Lectures", "Labs", "Exercises", "Projects"):
            continue
        print(f"  ...section {title}...")

        for link in links:
            link = link.strip()
            folder = folderNameForURL(link)
            if not RESCAN_ALL and link in skip:
                print(f"    ...skipping {folder}...")
                continue

            if folder is not None:
                print(f"    ...checking {folder}...")
                discovered = None
                os.makedirs(folder, exist_ok=True)
                fullLink = urllib.parse.urljoin(CALENDAR_URL, link)
                if title == 'Lectures':
                    discovered = findLectureURLs(folder, fullLink)
                    downloadMap.update(discovered)
                elif title == 'Labs':
                    discovered = findLabURLs(folder, fullLink)
                    downloadMap.update(discovered)
                elif title == 'Exercises':
                    discovered = findExercisesURLs(folder, fullLink)
                    downloadMap.update(discovered)
                elif title == 'Projects':
                    discovered = findProjectURLs(folder, fullLink)
                    downloadMap.update(discovered)

                if discovered is not None:
                    skip.append(link)

    print("  ...done.")

    # Save skip list to skip next time we run
    with open('skip.txt', 'w') as outputFile:
        for url in skip:
            outputFile.write(url + '\n')

    new = []
    for url, destination in downloadMap.items():
        if not os.path.exists(destination):
            new.append((url, destination))

    print(f"Found {len(new)} new items to download..")

    # TODO: Graceful failure and skip undoing here!
    for url, destination in new:
        print(f"Downloading {destination}...")
        with urllib.request.urlopen(url) as webInput:
            with open(destination, 'wb') as fileOutput:
                fileOutput.write(webInput.read())
            if url.endswith('.zip'):  # Also extract the zip file
                print("  ...extracting zip file.")
                extractor = zipfile.ZipFile(destination)
                extractor.extractall(path=os.path.split(destination)[0])
        print("  ...done.")


if __name__ == "__main__":
    main()
